import copy

# from wikidataintegrator.wdi_core import *

__author__ = 'Sebastian Burgstaller-Muehlbacher'
__license__ = 'AGPLv3'

prefix = '''
    PREFIX wd: <http://www.wikidata.org/entity/>
    PREFIX wdt: <http://www.wikidata.org/prop/direct/>
    PREFIX p: <http://www.wikidata.org/prop/>
    PREFIX ps: <http://www.wikidata.org/prop/statement/>
    PREFIX pq: <http://www.wikidata.org/prop/qualifier/>
    PREFIX skos:<http://www.w3.org/2004/02/skos/core#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    '''


class FastRunContainer(object):
    def __init__(self, base_data_type, engine, base_filter=None):
        self.prop_data = {}
        self.loaded_langs = {}
        self.statements = []
        self.base_filter = {}
        self.base_filter_string = ''
        self.prop_dt_map = {}
        self.current_qid = ''
        self.rev_lookup = {}
        self.base_data_type = base_data_type
        self.engine = engine

        if base_filter and any(base_filter):
            self.base_filter = base_filter

            for k, v in self.base_filter.items():
                if v:
                    self.base_filter_string += '?p p:{0}/ps:{0} wd:{1} . \n'.format(k, v)
                else:
                    self.base_filter_string += '?p p:{0}/ps:{0} ?zz . \n'.format(k)

    def check_data(self, data, append_props=None, cqid=None):
        del_props = set()
        data_props = set()
        if not append_props:
            append_props = []

        for x in data:
            if x.value and x.data_type:
                data_props.add(x.get_prop_nr())
        write_required = False
        match_sets = []
        for date in data:
            # skip to next if statement has no value or no data type defined, e.g. for deletion objects
            current_value = date.get_value()
            if not current_value and not date.data_type:
                del_props.add(date.get_prop_nr())
                continue

            prop_nr = date.get_prop_nr()

            if prop_nr not in self.prop_dt_map:
                self.prop_dt_map.update({prop_nr: FastRunContainer.get_prop_datatype(prop_nr=prop_nr,
                                                                                     engine=self.engine)})
                self.__query_data(prop_nr=prop_nr)

            # more sophisticated data types like dates and globe coordinates need special treatment here
            if self.prop_dt_map[prop_nr] == 'time':
                current_value = current_value[0]
            elif self.prop_dt_map[prop_nr] == 'wikibase-item':
                current_value = 'Q{}'.format(current_value)
            elif self.prop_dt_map[prop_nr] == 'globe-coordinate':
                write_required = True  # temporary workaround for handling globe coordinates

            if not __debug__:
                print(current_value)
            try:
                if not __debug__:
                    print(current_value)
                temp_set = set(self.rev_lookup[current_value])
            except KeyError:
                if not __debug__:
                    print('no matches for rev lookup')
                return True

            match_sets.append(temp_set)

        if cqid:
            matching_qids = {cqid}
        else:
            matching_qids = match_sets[0].intersection(*match_sets[1:])

        if not len(matching_qids) == 1:
            if not __debug__:
                print('no matches')
            return True

        qid = matching_qids.pop()
        self.current_qid = qid
        reconstructed_statements = []

        for prop_nr, dt in self.prop_data[qid].items():
            all_uids = set([x['s2'] for x in dt])
            q_props = set([x['pr'] for x in dt if 'pr' in x])
            for q_prop in q_props:
                if q_prop not in self.prop_dt_map:
                    self.prop_dt_map.update({q_prop: FastRunContainer.get_prop_datatype(prop_nr=q_prop,
                                                                                        engine=self.engine)})
            for uid in all_uids:
                qualifiers = [[x for x in self.base_data_type.__subclasses__() if x.DTYPE ==
                               self.prop_dt_map[y['pr']]][0](value=y['q'], prop_nr=y['pr'], is_qualifier=True)
                              for y in dt if y['s2'] == uid and 'q' in y]

                stmts = [[x for x in self.base_data_type.__subclasses__() if x.DTYPE ==
                          self.prop_dt_map[prop_nr]][0](value=y['v'], prop_nr=prop_nr, qualifiers=qualifiers)
                         for y in dt if y['s2'] == uid][0]

                reconstructed_statements.append(stmts)

        tmp_rs = copy.deepcopy(reconstructed_statements)

        # handle append properties
        for p in append_props:
            app_data = [x for x in data if x.get_prop_nr() == p]
            rec_app_data = [x for x in tmp_rs if x.get_prop_nr() == p]
            comp = [True for x in app_data for y in rec_app_data if x == y]
            if len(comp) != len(app_data):
                return True

        tmp_rs = [x for x in tmp_rs if x.get_prop_nr() not in append_props and x.get_prop_nr() in data_props]

        for date in data:
            # ensure that statements meant for deletion get handled properly
            reconst_props = set([x.get_prop_nr() for x in tmp_rs])
            if (not date.value or not date.data_type) and date.get_prop_nr() in reconst_props:
                if not __debug__:
                    print('returned from delete prop handling')
                return True
            elif not date.value or not date.data_type:
                # Ignore the deletion statements which are not in the reconstituted statements.
                continue

            if date.get_prop_nr() in append_props:
                continue
            bool_vec = [True if date == x and x.get_prop_nr() not in del_props else False for x in tmp_rs]

            if not __debug__:
                bool_vec = []

                print('-----------------------------------')
                for x in tmp_rs:

                    if date == x and x.get_prop_nr() not in del_props:
                        bool_vec.append(True)
                        print(x.get_prop_nr(), x.get_value(), [z.get_value() for z in x.get_qualifiers()])
                        print(date.get_prop_nr(), date.get_value(), [z.get_value() for z in date.get_qualifiers()])
                    else:
                        if x.get_prop_nr() == date.get_prop_nr():
                            print(x.get_prop_nr(), x.get_value(), [z.get_value() for z in x.get_qualifiers()])
                            print(date.get_prop_nr(), date.get_value(), [z.get_value() for z in date.get_qualifiers()])
                        bool_vec.append(False)

            if not any(bool_vec):
                if not __debug__:
                    print(len(bool_vec))
                    print('fast run failed at ', date.get_prop_nr())
                write_required = True
            else:
                tmp_rs.pop(bool_vec.index(True))

        if len(tmp_rs) > 0:
            if not __debug__:
                print('failed because not zero')
                for x in tmp_rs:
                    print('xxx', x.get_prop_nr(), x.get_value(), [z.get_value() for z in x.get_qualifiers()])
                print('failed because not zero--END')
            write_required = True

        return write_required

    def init_language_data(self, lang, lang_data_type):
        """
        Initialize language data store
        :param lang: language code
        :param lang_data_type: 'label', 'description' or 'aliases'
        :return: None
        """
        if lang not in self.loaded_langs:
            self.loaded_langs[lang] = {}

        if lang_data_type not in self.loaded_langs[lang]:
            self.loaded_langs[lang].update({lang_data_type: self.__query_lang(lang=lang,
                                                                              lang_data_type=lang_data_type)})

    def get_language_data(self, qid, lang, lang_data_type):
        """
        get language data for specified qid
        :param qid:
        :param lang: language code
        :param lang_data_type: 'label', 'description' or 'aliases'
        :return: list of strings
        """
        self.init_language_data(lang, lang_data_type)

        current_lang_data = self.loaded_langs[lang][lang_data_type]['results']['bindings']
        all_lang_strings = []
        for sresult in current_lang_data:
            if sresult['p']['value'].split('/')[-1] == qid:
                if 'label' in sresult:
                    all_lang_strings.append(sresult['label']['value'])
        if not all_lang_strings and lang_data_type in {'label', 'description'}:
            all_lang_strings = ['']
        return all_lang_strings

    def check_language_data(self, qid, lang_data, lang, lang_data_type):
        """
        Method to check if certain language data exists as a label, description or aliases
        :param lang_data: list of string values to check
        :type lang_data: list
        :param lang: language code
        :type lang: str
        :param lang_data_type: What kind of data is it? 'label', 'description' or 'aliases'?
        :return:
        """
        all_lang_strings = self.get_language_data(qid, lang, lang_data_type)

        for s in lang_data:
            if s not in all_lang_strings:
                print('fastrun failed at label: ', lang_data_type)
                return True

        return False

    def get_all_data(self):
        return self.prop_data

    def __query_data(self, prop_nr):
        query = '''
            #Tool: wdi_core fastrun
            select ?p ?q ?pr ?s2 ?v where {{
              {0}

              ?p p:{1} ?s2 .

              ?s2 ps:{1} ?v .
              OPTIONAL {{
                ?s2 ?pr ?q .
                FILTER(STRSTARTS(STR(?pr), "http://www.wikidata.org/prop/qualifier/"))
              }}

            }}
        '''.format(self.base_filter_string, prop_nr)

        # query for amount
        '''
        PREFIX wikibase: <http://wikiba.se/ontology#>
        SELECT ?p ?q ?pr ?s2 ?v ?psv ?amount ?upper_bound ?lower_bound ?unit WHERE {
            ?p p:P31/ps:P31 wd:Q11173 .

            ?p p:P2067 ?s2 .

            ?s2 ps:P2067 ?v .
            ?s2 <http://www.wikidata.org/prop/statement/value/P2067> ?psv .
            ?psv wikibase:quantityAmount ?amount .
            ?psv wikibase:quantityUpperBound ?upper_bound .
            ?psv wikibase:quantityLowerBound ?lower_bound .
            ?psv wikibase:quantityUnit ?unit .

            OPTIONAL {
                ?s2 ?pr ?q .
                FILTER(STRSTARTS(STR(?pr), "http://www.wikidata.org/prop/qualifier/"))
              }

            }
        '''

        # query for globe coordinate
        '''
        PREFIX wikibase: <http://wikiba.se/ontology#>
        SELECT ?p ?q ?pr ?s2 ?v ?psv ?geoLatitude ?geoLongitude ?geoGlobe ?geoPrecision WHERE {
            ?p p:P30/ps:P30 wd:Q46 .

            ?p p:P625 ?s2 .

            ?s2 ps:P625 ?v .
            ?s2 <http://www.wikidata.org/prop/statement/value/P625> ?psv .
            ?psv wikibase:geoLatitude ?geoLatitude .
            ?psv wikibase:geoLongitude ?geoLongitude .
            ?psv wikibase:geoGlobe ?geoGlobe .
            ?psv wikibase:geoPrecision ?geoPrecision .
            OPTIONAL {
                ?s2 ?pr ?q .
                FILTER(STRSTARTS(STR(?pr), "http://www.wikidata.org/prop/qualifier/"))
              }

            }
        '''

        if not __debug__:
            print(query)

        r = self.engine.execute_sparql_query(query=query, prefix=prefix)

        for i in r['results']['bindings']:
            i['p'] = i['p']['value'].split('/')[-1]
            if 's2' in i:
                i['s2'] = i['s2']['value'].split('/')[-1]
            if 'q' in i:
                i['q'] = i['q']['value'].split('/')[-1]
            if 'pr' in i:
                i['pr'] = i['pr']['value'].split('/')[-1]

            if 'v' in i:
                if i['v']['type'] == 'literal':
                    i['v'] = i['v']['value']
                elif i['v']['type'] == 'uri':
                    if 'www.wikidata.org/entity/' in i['v']['value']:
                        i['v'] = i['v']['value'].split('/')[-1]
                    else:
                        i['v'] = i['v']['value']

                # TODO: needs check for no-value and some-value sparql results return
                # for some-value, this json is being returned {'value': 't329541227', 'type': 'bnode'}
                if type(i['v']) is not dict and i['v'] in self.rev_lookup:
                    self.rev_lookup[i['v']].append(i['p'])
                else:
                    self.rev_lookup[i['v']] = [i['p']]

        if not __debug__:
            print('Length of reverse lookup table:', len(self.rev_lookup))

        for i in r['results']['bindings']:
            qid = i['p']
            if qid not in self.prop_data:
                self.prop_data[qid] = {prop_nr: []}
            if prop_nr not in self.prop_data[qid]:
                self.prop_data[qid].update({prop_nr: []})

            t = copy.deepcopy(i)
            del t['p']
            self.prop_data[qid][prop_nr].append(t)

        del r

    def __query_lang(self, lang, lang_data_type):
        """

        :param lang:
        :param lang_data_type:
        :return:
        """

        lang_data_type_dict = {
            'label': 'rdfs:label',
            'description': 'schema:description',
            'aliases': 'skos:altLabel'
        }

        query = '''
        #Tool: wdi_core fastrun
        SELECT ?p ?label WHERE {{
            {0}

            OPTIONAL {{
                ?p {1} ?label FILTER (lang(?label) = "{2}") .
            }}
        }}
        '''.format(self.base_filter_string, lang_data_type_dict[lang_data_type], lang)

        if not __debug__:
            print(query)

        return self.engine.execute_sparql_query(query=query, prefix=prefix)

    @staticmethod
    def get_prop_datatype(prop_nr, engine):
        item = engine(wd_item_id=prop_nr)
        return item.entity_metadata['datatype']



# coding:utf-8
"""
Function decoration for backoff and retry

This module provides function decorators which can be used to wrap a
function such that it will be retried until some condition is met. It
is meant to be of use when accessing unreliable resources with the
potential for intermittent failures i.e. network resources and external
APIs. Somewhat more generally, it may also be of use for dynamically
polling resources for externally generated content.

For examples and full documentation see the README at
https://github.com/litl/backoff
"""
from __future__ import unicode_literals

import functools
import operator
import logging
import random
import time
import traceback
import sys


# Use module-specific logger with a default null handler.
logger = logging.getLogger(__name__)

if sys.version_info < (2, 7, 0):  # pragma: no cover
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
    logger.addHandler(NullHandler())
else:
    logger.addHandler(logging.NullHandler())  # pragma: no cover

logger.setLevel(logging.ERROR)


def expo(base=2, factor=1, max_value=None):
    """Generator for exponential decay.

    Args:
        base: The mathematical base of the exponentiation operation
        factor: Factor to multiply the exponentation by.
        max_value: The maximum value to yield. Once the value in the
             true exponential sequence exceeds this, the value
             of max_value will forever after be yielded.
    """
    n = 0
    while True:
        a = factor * base ** n
        if max_value is None or a < max_value:
            yield a
            n += 1
        else:
            yield max_value


def fibo(max_value=None):
    """Generator for fibonaccial decay.

    Args:
        max_value: The maximum value to yield. Once the value in the
             true fibonacci sequence exceeds this, the value
             of max_value will forever after be yielded.
    """
    a = 1
    b = 1
    while True:
        if max_value is None or a < max_value:
            yield a
            a, b = b, a + b
        else:
            yield max_value


def constant(interval=1):
    """Generator for constant intervals.

    Args:
        interval: The constant value in seconds to yield.
    """
    while True:
        yield interval


def random_jitter(value):
    """Jitter the value a random number of milliseconds.

    This adds up to 1 second of additional time to the original value.
    Prior to backoff version 1.2 this was the default jitter behavior.

    Args:
        value: The unadulterated backoff value.
    """
    return value + random.random()


def full_jitter(value):
    """Jitter the value across the full range (0 to value).

    This corresponds to the "Full Jitter" algorithm specified in the
    AWS blog's post on the performance of various jitter algorithms.
    (http://www.awsarchitectureblog.com/2015/03/backoff.html)

    Args:
        value: The unadulterated backoff value.
    """
    return random.uniform(0, value)


def on_predicate(wait_gen,
                 predicate=operator.not_,
                 max_tries=None,
                 jitter=full_jitter,
                 on_success=None,
                 on_backoff=None,
                 on_giveup=None,
                 **wait_gen_kwargs):
    """Returns decorator for backoff and retry triggered by predicate.

    Args:
        wait_gen: A generator yielding successive wait times in
            seconds.
        predicate: A function which when called on the return value of
            the target function will trigger backoff when considered
            truthily. If not specified, the default behavior is to
            backoff on falsey return values.
        max_tries: The maximum number of attempts to make before giving
            up. In the case of failure, the result of the last attempt
            will be returned.  The default value of None means their
            is no limit to the number of tries.
        jitter: A function of the value yielded by wait_gen returning
            the actual time to wait. This distributes wait times
            stochastically in order to avoid timing collisions across
            concurrent clients. Wait times are jittered by default
            using the full_jitter function. Jittering may be disabled
            altogether by passing jitter=None.
        on_success: Callable (or iterable of callables) with a unary
            signature to be called in the event of success. The
            parameter is a dict containing details about the invocation.
        on_backoff: Callable (or iterable of callables) with a unary
            signature to be called in the event of a backoff. The
            parameter is a dict containing details about the invocation.
        on_giveup: Callable (or iterable of callables) with a unary
            signature to be called in the event that max_tries
            is exceeded.  The parameter is a dict containing details
            about the invocation.
        **wait_gen_kwargs: Any additional keyword args specified will be
            passed to wait_gen when it is initialized.  Any callable
            args will first be evaluated and their return values passed.
            This is useful for runtime configuration.
    """
    success_hdlrs = _handlers(on_success)
    backoff_hdlrs = _handlers(on_backoff, _log_backoff)
    giveup_hdlrs = _handlers(on_giveup, _log_giveup)

    def decorate(target):

        @functools.wraps(target)
        def retry(*args, **kwargs):
            # change names because python 2.x doesn't have nonlocal
            max_tries_ = _maybe_call(max_tries)

            # there are no dictionary comprehensions in python 2.6
            wait = wait_gen(**dict((k, _maybe_call(v))
                                   for k, v in wait_gen_kwargs.items()))

            tries = 0
            while True:
                tries += 1
                ret = target(*args, **kwargs)
                if predicate(ret):
                    if tries == max_tries_:
                        for hdlr in giveup_hdlrs:
                            hdlr({'target': target,
                                  'args': args,
                                  'kwargs': kwargs,
                                  'tries': tries,
                                  'value': ret})
                        break

                    value = next(wait)
                    try:
                        if jitter is not None:
                            seconds = jitter(value)
                        else:
                            seconds = value
                    except TypeError:
                        # support deprecated nullary jitter function signature
                        # which returns a delta rather than a jittered value
                        seconds = value + jitter()

                    for hdlr in backoff_hdlrs:
                        hdlr({'target': target,
                              'args': args,
                              'kwargs': kwargs,
                              'tries': tries,
                              'value': ret,
                              'wait': seconds})

                    time.sleep(seconds)
                    continue
                else:
                    for hdlr in success_hdlrs:
                        hdlr({'target': target,
                              'args': args,
                              'kwargs': kwargs,
                              'tries': tries,
                              'value': ret})
                    break

            return ret

        return retry

    # Return a function which decorates a target with a retry loop.
    return decorate


def on_exception(wait_gen,
                 exception,
                 max_tries=None,
                 jitter=full_jitter,
                 giveup=lambda e: False,
                 on_success=None,
                 on_backoff=None,
                 on_giveup=None,
                 **wait_gen_kwargs):
    """Returns decorator for backoff and retry triggered by exception.

    Args:
        wait_gen: A generator yielding successive wait times in
            seconds.
        exception: An exception type (or tuple of types) which triggers
            backoff.
        max_tries: The maximum number of attempts to make before giving
            up. Once exhausted, the exception will be allowed to escape.
            The default value of None means their is no limit to the
            number of tries. If a callable is passed, it will be
            evaluated at runtime and its return value used.
        jitter: A function of the value yielded by wait_gen returning
            the actual time to wait. This distributes wait times
            stochastically in order to avoid timing collisions across
            concurrent clients. Wait times are jittered by default
            using the full_jitter function. Jittering may be disabled
            altogether by passing jitter=None.
        giveup: Function accepting an exception instance and
            returning whether or not to give up. Optional. The default
            is to always continue.
        on_success: Callable (or iterable of callables) with a unary
            signature to be called in the event of success. The
            parameter is a dict containing details about the invocation.
        on_backoff: Callable (or iterable of callables) with a unary
            signature to be called in the event of a backoff. The
            parameter is a dict containing details about the invocation.
        on_giveup: Callable (or iterable of callables) with a unary
            signature to be called in the event that max_tries
            is exceeded.  The parameter is a dict containing details
            about the invocation.
        **wait_gen_kwargs: Any additional keyword args specified will be
            passed to wait_gen when it is initialized.  Any callable
            args will first be evaluated and their return values passed.
            This is useful for runtime configuration.
    """
    success_hdlrs = _handlers(on_success)
    backoff_hdlrs = _handlers(on_backoff, _log_backoff)
    giveup_hdlrs = _handlers(on_giveup, _log_giveup)

    def decorate(target):

        @functools.wraps(target)
        def retry(*args, **kwargs):
            # change names because python 2.x doesn't have nonlocal
            max_tries_ = _maybe_call(max_tries)

            # there are no dictionary comprehensions in python 2.6
            wait = wait_gen(**dict((k, _maybe_call(v))
                                   for k, v in wait_gen_kwargs.items()))

            tries = 0
            while True:
                try:
                    tries += 1
                    ret = target(*args, **kwargs)
                except exception as e:
                    if giveup(e) or tries == max_tries_:
                        for hdlr in giveup_hdlrs:
                            hdlr({'target': target,
                                  'args': args,
                                  'kwargs': kwargs,
                                  'tries': tries})
                        raise

                    value = next(wait)
                    try:
                        if jitter is not None:
                            seconds = jitter(value)
                        else:
                            seconds = value
                    except TypeError:
                        # support deprecated nullary jitter function signature
                        # which returns a delta rather than a jittered value
                        seconds = value + jitter()

                    for hdlr in backoff_hdlrs:
                        hdlr({'target': target,
                              'args': args,
                              'kwargs': kwargs,
                              'tries': tries,
                              'wait': seconds})

                    time.sleep(seconds)
                else:
                    for hdlr in success_hdlrs:
                        hdlr({'target': target,
                              'args': args,
                              'kwargs': kwargs,
                              'tries': tries})

                    return ret

        return retry

    # Return a function which decorates a target with a retry loop.
    return decorate


# Create default handler list from keyword argument
def _handlers(hdlr, default=None):
    defaults = [default] if default is not None else []

    if hdlr is None:
        return defaults

    if hasattr(hdlr, '__iter__'):
        return defaults + list(hdlr)

    return defaults + [hdlr]


# Evaluate arg that can be either a fixed value or a callable.
def _maybe_call(f, *args, **kwargs):
    return f(*args, **kwargs) if callable(f) else f


# Default backoff handler
def _log_backoff(details):
    fmt = "Backing off {0}(...) for {1:.1f}s"
    msg = fmt.format(details['target'].__name__, details['wait'])

    exc_typ, exc, _ = sys.exc_info()
    if exc is not None:
        exc_fmt = traceback.format_exception_only(exc_typ, exc)[-1]
        msg = "{0} ({1})".format(msg, exc_fmt.rstrip("\n"))
        logger.error(msg)
    else:
        msg = "{0} ({1})".format(msg, details['value'])
        logger.info(msg)


# Default giveup handler
def _log_giveup(details):
    fmt = "Giving up {0}(...) after {1} tries"
    msg = fmt.format(details['target'].__name__, details['tries'])

    exc_typ, exc, _ = sys.exc_info()
    if exc is not None:
        exc_fmt = traceback.format_exception_only(exc_typ, exc)[-1]
        msg = "{0} ({1})".format(msg, exc_fmt.rstrip("\n"))
    else:
        msg = "{0} ({1})".format(msg, details['value'])

    logger.error(msg)



#!usr/bin/env python
# -*- coding: utf-8 -*-

'''
Authors: 
  Sebastian Burgstaller (sebastian.burgstaller' at 'gmail.com
  Andra Waagmeester (andra' at ' micelio.be)

This file is part of ProteinBoxBot.

ProteinBoxBot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

ProteinBoxBot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with ProteinBoxBot.  If not, see <http://www.gnu.org/licenses/>.
'''


# from raven import Client

def getSentryKey():
    return '<Sentry key to be obtained from https://getsentry.com/welcome/>'


def getWikiDataUser():
    return 'ProteinBoxBot'


def getWikiDataPassword():
    return '<WikiData Password>'


def getMySQLHost():
    return '127.0.0.1'


def getMySQLUser():
    return 'PBB'


def getMySQLPW():
    return ''



import time

import requests

from wikidataintegrator.backoff.wdi_backoff import wdi_backoff, get_config

__author__ = 'Andra Waagmeester and Sebastian Burgstaller'
__license__ = 'AGPLv3'

"""
Login class for Wikidata. Takes username and password and stores the session cookies and edit tokens.
"""


class WDLogin(object):
    """
    A class which handles the login to Wikidata and the generation of edit-tokens
    """

    @wdi_backoff()
    def __init__(self, user, pwd, server='www.wikidata.org', token_renew_period=1800, use_clientlogin=False):
        """
        constructor
        :param user: the username which should be used for the login
        :param pwd: the password which should be used for the login
        :param server: the wikimedia server the login should be made to
        :param token_renew_period: Seconds after which a new token should be requested from the Wikidata server
        :type token_renew_period: int
        :param use_clientlogin: use authmanager based login method instead of standard login.
            For 3rd party data consumer, e.g. web clients
        :type bool
        :return: None
        """
        if server is not None:
            self.server = server
        self.base_url = 'https://{}/w/api.php'.format(self.server)
        self.s = requests.Session()

        self.edit_token = ''
        self.instantiation_time = time.time()
        self.token_renew_period = token_renew_period

        if use_clientlogin:
            params = {
                'action': 'query',
                'format': 'json',
                'meta': 'authmanagerinfo',
                'amisecuritysensitiveoperation': '',
                'amirequestsfor': 'login'
            }

            self.s.get(self.base_url, params=params)

            params2 = {
                'action': 'query',
                'format': 'json',
                'meta': 'tokens',
                'type': 'login'
            }
            login_token = self.s.get(self.base_url, params=params2).json()['query']['tokens']['logintoken']

            data = {
                'action': 'clientlogin',
                'format': 'json',
                'username': user,
                'password': pwd,
                'logintoken': login_token,
                'loginreturnurl': 'http://example.org/'
            }

            login_result = self.s.post(self.base_url, data=data).json()
            print(login_result)

            if login_result['clientlogin']['status'] == 'FAIL':
                raise ValueError('Login FAILED')
        else:
            params = {
                'action': 'login',
                'lgname': user,
                'lgpassword': pwd,
                'format': 'json'
            }

            # get login token
            login_token = self.s.post(self.base_url, data=params).json()['login']['token']

            # do the login using the login token
            params.update({'lgtoken': login_token})
            r = self.s.post(self.base_url, data=params).json()

            if r['login']['result'] != 'Success':
                print('login failed:', r['login']['reason'])
                raise ValueError('login FAILED!!')
            else:
                print('Successfully logged in as', r['login']['lgusername'])

        self.generate_edit_credentials()

    def generate_edit_credentials(self):
        """
        request an edit token and update the cookie_jar in order to add the session cookie
        :return: Returns a json with all relevant cookies, aka cookie jar
        """
        params = {
            'action': 'query',
            'meta': 'tokens',
            'format': 'json'
        }
        response = self.s.get(self.base_url, params=params)
        self.edit_token = response.json()['query']['tokens']['csrftoken']

        return self.s.cookies

    def get_edit_cookie(self):
        """
        Can be called in order to retrieve the cookies from an instance of WDLogin
        :return: Returns a json with all relevant cookies, aka cookie jar
        """
        if (time.time() - self.instantiation_time) > self.token_renew_period:
            self.generate_edit_credentials()
            self.instantiation_time = time.time()

        return self.s.cookies

    def get_edit_token(self):
        """
        Can be called in order to retrieve the edit token from an instance of WDLogin
        :return: returns the edit token
        """
        if not self.edit_token or (time.time() - self.instantiation_time) > self.token_renew_period:
            self.generate_edit_credentials()
            self.instantiation_time = time.time()

        return self.edit_token

    def get_session(self):
        """
        returns the requests session object used for the login.
        :return: Object of type requests.Session()
        """
        return self.s



"""
Config global options
Options can be changed at run time. See tests/test_backoff.py for usage example

Options:
BACKOFF_MAX_TRIES: maximum number of times to retry failed request to wikidata endpoint.
        Default: None (retry indefinitely)
        To disable retry, set value to 1
BACKOFF_MAX_VALUE: maximum number of seconds to wait before retrying. wait time will increase to this number
        Default: 3600 (one hour)
"""

config = {'BACKOFF_MAX_TRIES': None,
          'BACKOFF_MAX_VALUE': 3600}


# coding:utf-8

import backoff
import collections
import functools
import pytest
import random


def test_full_jitter():
    for input in range(100):
        for i in range(100):
            jitter = backoff.full_jitter(input)
        assert jitter >= 0
        assert jitter <= input


def test_expo():
    gen = backoff.expo()
    for i in range(9):
        assert 2 ** i == next(gen)


def test_expo_base3():
    gen = backoff.expo(base=3)
    for i in range(9):
        assert 3 ** i == next(gen)


def test_expo_factor3():
    gen = backoff.expo(factor=3)
    for i in range(9):
        assert 3 * 2 ** i == next(gen)


def test_expo_base3_factor5():
    gen = backoff.expo(base=3, factor=5)
    for i in range(9):
        assert 5 * 3 ** i == next(gen)


def test_expo_max_value():
    gen = backoff.expo(max_value=2 ** 4)
    expected = [1, 2, 4, 8, 16, 16, 16]
    for expect in expected:
        assert expect == next(gen)


def test_fibo():
    gen = backoff.fibo()
    expected = [1, 1, 2, 3, 5, 8, 13]
    for expect in expected:
        assert expect == next(gen)


def test_fibo_max_value():
    gen = backoff.fibo(max_value=8)
    expected = [1, 1, 2, 3, 5, 8, 8, 8]
    for expect in expected:
        assert expect == next(gen)


def test_constant():
    gen = backoff.constant(interval=3)
    for i in range(9):
        assert 3 == next(gen)


def test_on_predicate(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    @backoff.on_predicate(backoff.expo)
    def return_true(log, n):
        val = (len(log) == n - 1)
        log.append(val)
        return val

    log = []
    ret = return_true(log, 3)
    assert ret is True
    assert 3 == len(log)


def test_on_predicate_max_tries(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    @backoff.on_predicate(backoff.expo, jitter=None, max_tries=3)
    def return_true(log, n):
        val = (len(log) == n)
        log.append(val)
        return val

    log = []
    ret = return_true(log, 10)
    assert ret is False
    assert 3 == len(log)


def test_on_exception(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    @backoff.on_exception(backoff.expo, KeyError)
    def keyerror_then_true(log, n):
        if len(log) == n:
            return True
        e = KeyError()
        log.append(e)
        raise e

    log = []
    assert keyerror_then_true(log, 3) is True
    assert 3 == len(log)


def test_on_exception_tuple(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    @backoff.on_exception(backoff.expo, (KeyError, ValueError))
    def keyerror_valueerror_then_true(log):
        if len(log) == 2:
            return True
        if len(log) == 0:
            e = KeyError()
        if len(log) == 1:
            e = ValueError()
        log.append(e)
        raise e

    log = []
    assert keyerror_valueerror_then_true(log) is True
    assert 2 == len(log)
    assert isinstance(log[0], KeyError)
    assert isinstance(log[1], ValueError)


def test_on_exception_max_tries(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    @backoff.on_exception(backoff.expo, KeyError, jitter=None, max_tries=3)
    def keyerror_then_true(log, n, foo=None):
        if len(log) == n:
            return True
        e = KeyError()
        log.append(e)
        raise e

    log = []
    with pytest.raises(KeyError):
        keyerror_then_true(log, 10, foo="bar")

    assert 3 == len(log)


def _test_invoc_repr():
    def func(a, b, c=None):
        pass

    assert "func(a, b, c=c)" == backoff._invoc_repr((func,
                                                     ["a", "b"],
                                                     {"c": "c"}))
    assert "func(c=c)" == backoff._invoc_repr((func, [], {"c": "c"}))
    assert "func(a, b)" == backoff._invoc_repr((func, ["a", "b"], {}))
    assert u"func(ユニコーン, ア=あ)" == \
        backoff._invoc_repr((func, [u"ユニコーン"], {u"ア": u"あ"}))

    # tuple args caused a string formatting exception
    assert "func((1, 2, 3))" == backoff._invoc_repr((func, [(1, 2, 3)], {}))


# create event handler which log their invocations to a dict
def _log_hdlrs():
    log = collections.defaultdict(list)

    def log_hdlr(event, details):
        log[event].append(details)

    log_success = functools.partial(log_hdlr, 'success')
    log_backoff = functools.partial(log_hdlr, 'backoff')
    log_giveup = functools.partial(log_hdlr, 'giveup')

    return log, log_success, log_backoff, log_giveup


# decorator that that saves the target as
# an attribute of the decorated function
def _save_target(f):
    f._target = f
    return f


def test_on_exception_success_random_jitter(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    log, log_success, log_backoff, log_giveup = _log_hdlrs()

    @backoff.on_exception(backoff.expo,
                          Exception,
                          on_success=log_success,
                          on_backoff=log_backoff,
                          on_giveup=log_giveup,
                          jitter=backoff.random_jitter,
                          factor=0.5)
    @_save_target
    def succeeder(*args, **kwargs):
        # succeed after we've backed off twice
        if len(log['backoff']) < 2:
            raise ValueError("catch me")

    succeeder(1, 2, 3, foo=1, bar=2)

    # we try 3 times, backing off twice before succeeding
    assert len(log['success']) == 1
    assert len(log['backoff']) == 2
    assert len(log['giveup']) == 0

    for i in range(2):
        details = log['backoff'][i]
        assert details['wait'] >= 0.5 * 2 ** i


def test_on_exception_success_full_jitter(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    log, log_success, log_backoff, log_giveup = _log_hdlrs()

    @backoff.on_exception(backoff.expo,
                          Exception,
                          on_success=log_success,
                          on_backoff=log_backoff,
                          on_giveup=log_giveup,
                          jitter=backoff.full_jitter,
                          factor=0.5)
    @_save_target
    def succeeder(*args, **kwargs):
        # succeed after we've backed off twice
        if len(log['backoff']) < 2:
            raise ValueError("catch me")

    succeeder(1, 2, 3, foo=1, bar=2)

    # we try 3 times, backing off twice before succeeding
    assert len(log['success']) == 1
    assert len(log['backoff']) == 2
    assert len(log['giveup']) == 0

    for i in range(2):
        details = log['backoff'][i]
        assert details['wait'] <= 0.5 * 2 ** i


def test_on_exception_success():
    log, log_success, log_backoff, log_giveup = _log_hdlrs()

    @backoff.on_exception(backoff.constant,
                          Exception,
                          on_success=log_success,
                          on_backoff=log_backoff,
                          on_giveup=log_giveup,
                          jitter=lambda: 0,
                          interval=0)
    @_save_target
    def succeeder(*args, **kwargs):
        # succeed after we've backed off twice
        if len(log['backoff']) < 2:
            raise ValueError("catch me")

    succeeder(1, 2, 3, foo=1, bar=2)

    # we try 3 times, backing off twice before succeeding
    assert len(log['success']) == 1
    assert len(log['backoff']) == 2
    assert len(log['giveup']) == 0

    for i in range(2):
        details = log['backoff'][i]
        assert details == {'args': (1, 2, 3),
                           'kwargs': {'foo': 1, 'bar': 2},
                           'target': succeeder._target,
                           'tries': i + 1,
                           'wait': 0}

    details = log['success'][0]
    assert details == {'args': (1, 2, 3),
                       'kwargs': {'foo': 1, 'bar': 2},
                       'target': succeeder._target,
                       'tries': 3}


def test_on_exception_giveup():
    log, log_success, log_backoff, log_giveup = _log_hdlrs()

    @backoff.on_exception(backoff.constant,
                          ValueError,
                          on_success=log_success,
                          on_backoff=log_backoff,
                          on_giveup=log_giveup,
                          max_tries=3,
                          jitter=lambda: 0,
                          interval=0)
    @_save_target
    def exceptor(*args, **kwargs):
        raise ValueError("catch me")

    with pytest.raises(ValueError):
        exceptor(1, 2, 3, foo=1, bar=2)

    # we try 3 times, backing off twice and giving up once
    assert len(log['success']) == 0
    assert len(log['backoff']) == 2
    assert len(log['giveup']) == 1

    details = log['giveup'][0]
    assert details == {'args': (1, 2, 3),
                       'kwargs': {'foo': 1, 'bar': 2},
                       'target': exceptor._target,
                       'tries': 3}


def test_on_exception_giveup_predicate(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    def on_baz(e):
        return str(e) == "baz"

    vals = ["baz", "bar", "foo"]

    @backoff.on_exception(backoff.constant,
                          ValueError,
                          giveup=on_baz)
    def foo_bar_baz():
        raise ValueError(vals.pop())

    with pytest.raises(ValueError):
        foo_bar_baz()

    assert not vals


def test_on_predicate_success():
    log, log_success, log_backoff, log_giveup = _log_hdlrs()

    @backoff.on_predicate(backoff.constant,
                          on_success=log_success,
                          on_backoff=log_backoff,
                          on_giveup=log_giveup,
                          jitter=lambda: 0,
                          interval=0)
    @_save_target
    def success(*args, **kwargs):
        # succeed after we've backed off twice
        return len(log['backoff']) == 2

    success(1, 2, 3, foo=1, bar=2)

    # we try 3 times, backing off twice before succeeding
    assert len(log['success']) == 1
    assert len(log['backoff']) == 2
    assert len(log['giveup']) == 0

    for i in range(2):
        details = log['backoff'][i]
        assert details == {'args': (1, 2, 3),
                           'kwargs': {'foo': 1, 'bar': 2},
                           'target': success._target,
                           'tries': i + 1,
                           'value': False,
                           'wait': 0}

    details = log['success'][0]
    assert details == {'args': (1, 2, 3),
                       'kwargs': {'foo': 1, 'bar': 2},
                       'target': success._target,
                       'tries': 3,
                       'value': True}


def test_on_predicate_giveup():
    log, log_success, log_backoff, log_giveup = _log_hdlrs()

    @backoff.on_predicate(backoff.constant,
                          on_success=log_success,
                          on_backoff=log_backoff,
                          on_giveup=log_giveup,
                          max_tries=3,
                          jitter=lambda: 0,
                          interval=0)
    @_save_target
    def emptiness(*args, **kwargs):
        pass

    emptiness(1, 2, 3, foo=1, bar=2)

    # we try 3 times, backing off twice and giving up once
    assert len(log['success']) == 0
    assert len(log['backoff']) == 2
    assert len(log['giveup']) == 1

    details = log['giveup'][0]
    assert details == {'args': (1, 2, 3),
                       'kwargs': {'foo': 1, 'bar': 2},
                       'target': emptiness._target,
                       'tries': 3,
                       'value': None}


def test_on_predicate_iterable_handlers():
    hdlrs = [_log_hdlrs() for _ in range(3)]

    @backoff.on_predicate(backoff.constant,
                          on_success=(h[1] for h in hdlrs),
                          on_backoff=(h[2] for h in hdlrs),
                          on_giveup=(h[3] for h in hdlrs),
                          max_tries=3,
                          jitter=lambda: 0,
                          interval=0)
    @_save_target
    def emptiness(*args, **kwargs):
        pass

    emptiness(1, 2, 3, foo=1, bar=2)

    for i in range(3):
        assert len(hdlrs[i][0]['success']) == 0
        assert len(hdlrs[i][0]['backoff']) == 2
        assert len(hdlrs[i][0]['giveup']) == 1

        details = hdlrs[i][0]['giveup'][0]
        assert details == {'args': (1, 2, 3),
                           'kwargs': {'foo': 1, 'bar': 2},
                           'target': emptiness._target,
                           'tries': 3,
                           'value': None}


# To maintain backward compatibility,
# on_predicate should support 0-argument jitter function.
def test_on_exception_success_0_arg_jitter(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)
    monkeypatch.setattr('random.random', lambda: 0)

    log, log_success, log_backoff, log_giveup = _log_hdlrs()

    @backoff.on_exception(backoff.constant,
                          Exception,
                          on_success=log_success,
                          on_backoff=log_backoff,
                          on_giveup=log_giveup,
                          jitter=random.random,
                          interval=0)
    @_save_target
    def succeeder(*args, **kwargs):
        # succeed after we've backed off twice
        if len(log['backoff']) < 2:
            raise ValueError("catch me")

    succeeder(1, 2, 3, foo=1, bar=2)

    # we try 3 times, backing off twice before succeeding
    assert len(log['success']) == 1
    assert len(log['backoff']) == 2
    assert len(log['giveup']) == 0

    for i in range(2):
        details = log['backoff'][i]
        assert details == {'args': (1, 2, 3),
                           'kwargs': {'foo': 1, 'bar': 2},
                           'target': succeeder._target,
                           'tries': i + 1,
                           'wait': 0}

    details = log['success'][0]
    assert details == {'args': (1, 2, 3),
                       'kwargs': {'foo': 1, 'bar': 2},
                       'target': succeeder._target,
                       'tries': 3}


# To maintain backward compatibility,
# on_predicate should support 0-argument jitter function.
def test_on_predicate_success_0_arg_jitter(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)
    monkeypatch.setattr('random.random', lambda: 0)

    log, log_success, log_backoff, log_giveup = _log_hdlrs()

    @backoff.on_predicate(backoff.constant,
                          on_success=log_success,
                          on_backoff=log_backoff,
                          on_giveup=log_giveup,
                          jitter=random.random,
                          interval=0)
    @_save_target
    def success(*args, **kwargs):
        # succeed after we've backed off twice
        return len(log['backoff']) == 2

    success(1, 2, 3, foo=1, bar=2)

    # we try 3 times, backing off twice before succeeding
    assert len(log['success']) == 1
    assert len(log['backoff']) == 2
    assert len(log['giveup']) == 0

    for i in range(2):
        details = log['backoff'][i]
        assert details == {'args': (1, 2, 3),
                           'kwargs': {'foo': 1, 'bar': 2},
                           'target': success._target,
                           'tries': i + 1,
                           'value': False,
                           'wait': 0}

    details = log['success'][0]
    assert details == {'args': (1, 2, 3),
                       'kwargs': {'foo': 1, 'bar': 2},
                       'target': success._target,
                       'tries': 3,
                       'value': True}


def test_on_exception_callable_max_tries(monkeypatch):
    monkeypatch.setattr('time.sleep', lambda x: None)

    def lookup_max_tries():
        return 3

    log = []

    @backoff.on_exception(backoff.constant,
                          ValueError,
                          max_tries=lookup_max_tries)
    def exceptor():
        log.append(True)
        raise ValueError()

    with pytest.raises(ValueError):
        exceptor()

    assert len(log) == 3


def test_on_exception_callable_gen_kwargs():

    def lookup_foo():
        return "foo"

    def wait_gen(foo=None, bar=None):
        assert foo == "foo"
        assert bar == "bar"

        while True:
            yield 0

    @backoff.on_exception(wait_gen,
                          ValueError,
                          max_tries=2,
                          foo=lookup_foo,
                          bar="bar")
    def exceptor():
        raise ValueError("aah")

    with pytest.raises(ValueError):
        exceptor()



# coding:utf-8

import backoff

from distutils import core


classifiers = ['Development Status :: 5 - Production/Stable',
               'Intended Audience :: Developers',
               'Programming Language :: Python',
               'License :: OSI Approved :: MIT License',
               'Natural Language :: English',
               'Operating System :: OS Independent',
               'Programming Language :: Python',
               'Programming Language :: Python :: 2.6',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: Implementation',
               'Programming Language :: Python :: Implementation :: CPython',
               'Topic :: Internet :: WWW/HTTP',
               'Topic :: Software Development :: Libraries :: Python Modules',
               'Topic :: Utilities']


def readme():
    with open("README.rst", "r") as infile:
        return infile.read()


core.setup(name='backoff',
           version='1.3.2',
           description="Function decoration for backoff and retry",
           long_description=readme(),
           py_modules=['backoff'],
           author="Bob Green",
           author_email="rgreen@goscoutgo.com",
           keywords = "backoff function decorator",
           url="https://github.com/litl/backoff",
           download_url="https://github.com/litl/backoff/tarball/v1.3.2",
           license="MIT",
           classifiers=classifiers)



import datetime
import json
from collections import defaultdict, Counter
from time import gmtime, strftime

import requests
from dateutil import parser as du

from . import wdi_core
from .wdi_core import WDItemEngine, WDApiError


class Release(object):
    """
    Create a release item
    Example: Ensembl Release 86 Q27613766

    A relase item has the following required properties:
        title: **user_defined**
        description: **user_defined**
        instance of (P31): edition (Q3331189)
        edition or translation of (P629): **user_defined** (wd_item)
        edition number (P393): **user_defined** (str)
    Optional properties:
        archive URL (P1065): (str)
        publication date (P577): (wd_date)

    Example usage:
    r = PBB_Helpers.Release("Ensembl Release 85", "Release 85 of Ensembl", "85", edition_of="Ensembl",
                            archive_url="http://jul2016.archive.ensembl.org/",
                            pub_date='+2016-07-01T00:00:00Z', date_precision=10)
    r.create(login)

    """

    _release_cache = defaultdict(dict)
    _database_cache = dict()

    def __init__(self, title, description, edition, edition_of=None, edition_of_wdid=None, archive_url=None,
                 pub_date=None, date_precision=11):
        """

        :param title: title of release item
        :type title: str
        :param description: description of release item
        :type description: str
        :param edition: edition number or unique identifier for the release
        :type edition: str
        :param edition_of: name of database. database wdid will automatically be looked up. Must pass either edition_of or edition_of_wdid
        :type edition_of: str
        :param edition_of_wdid: wikidata id of database
        :type edition_of_wdid: str
        :param archive_url: (optional)
        :type archive_url: str
        :param pub_date: (optional) Datetime will be converted to str
        :type pub_date: str or datetime
        :param date_precision: (optional) passed to PBB_Core.WDTime as is. default is 11 (day)
        :type date_precision: int
        """
        self.title = title
        self.description = description
        self.edition = str(edition)
        self.archive_url = archive_url
        if isinstance(pub_date, datetime.date):
            self.pub_date = pub_date.strftime('+%Y-%m-%dT%H:%M:%SZ')
        else:
            self.pub_date = pub_date
        self.date_precision = date_precision

        if (edition_of is None and edition_of_wdid is None) or (edition_of and edition_of_wdid):
            raise ValueError("must provide either edition_of or edition_of_wdid")
        if edition_of:
            self.edition_of_wdid = self.lookup_database(edition_of)
        else:
            self.edition_of_wdid = edition_of_wdid

        self.statements = None
        self.make_statements()

    @classmethod
    def lookup_database(cls, edition_of):
        if edition_of in cls._database_cache:
            return cls._database_cache[edition_of]

        query = """SELECT ?item ?itemLabel WHERE {
                    ?item wdt:P31 wd:Q4117139 .
                    SERVICE wikibase:label { bd:serviceParam wikibase:language "en" }
                }"""
        results = WDItemEngine.execute_sparql_query(query)
        db_item_map = {x['itemLabel']['value']: x['item']['value'].replace('http://www.wikidata.org/entity/', '') for x
                       in results['results']['bindings']}
        if edition_of not in db_item_map:
            raise ValueError("Database {} not found in wikidata. Please provide edition_of_wdid".format(edition_of))

        cls._database_cache[edition_of] = db_item_map[edition_of]
        return db_item_map[edition_of]

    def make_statements(self):
        s = []
        # instance of edition
        s.append(wdi_core.WDItemID('Q3331189', 'P31'))
        # edition or translation of
        s.append(wdi_core.WDItemID(self.edition_of_wdid, 'P629'))
        # edition number
        s.append(wdi_core.WDString(self.edition, 'P393'))

        if self.archive_url:
            s.append(wdi_core.WDUrl(self.archive_url, 'P1065'))

        if self.pub_date:
            s.append(wdi_core.WDTime(self.pub_date, 'P577', precision=self.date_precision))

        self.statements = s

    def get_or_create(self, login=None):

        # check in cache
        if self.edition_of_wdid in self._release_cache and self.edition in self._release_cache[self.edition_of_wdid]:
            return self._release_cache[self.edition_of_wdid][self.edition]

        # check in wikidata
        edition_dict = id_mapper("P393", (("P629", self.edition_of_wdid), ("P31", "Q3331189")))
        if edition_dict and self.edition in edition_dict:
            # add to cache
            self._release_cache[self.edition_of_wdid][self.edition] = edition_dict[self.edition]
            return edition_dict[self.edition]

        # create new
        if login is None:
            raise ValueError("login required to create item")
        item = wdi_core.WDItemEngine(item_name=self.title, data=self.statements, domain="release")
        item.set_label(self.title)
        item.set_description(description=self.description, lang='en')
        write_success = try_write(item, self.edition + "|" + self.edition_of_wdid, 'P393|P629', login)
        if write_success:
            # add to cache
            self._release_cache[self.edition_of_wdid][self.edition] = item.wd_item_id
            return item.wd_item_id
        else:
            raise write_success


class PubmedItem(object):
    """
    Get or create a wikidata item stub given a pubmed ID

    Usage:
    PubmedStub(22674334).get_or_create(login)

    Data source: https://www.mediawiki.org/wiki/Citoid/API
    example:
    https://citoid.wikimedia.org/api?search=https://www.ncbi.nlm.nih.gov/pubmed/13054692&format=zotero

    wikidata Item format
    general guidelines: https://www.wikidata.org/wiki/Help:Sources
    more specific to journal articles: https://github.com/mitar/bib2wikidata/issues/1#issuecomment-50998834
    modelling after: https://www.wikidata.org/wiki/Q21093381 because it has both authors and author name strings

    This will not add authors, it will only add author name strings, as I do not habve enough information to assume
    they are the same as author items that exist.

    """

    _cache = dict()
    PROPS = {
        'instance of': 'P31',
        'title': 'P1476',
        'published in': 'P1433',
        'volume': 'P478',
        'publication date': 'P577',
        'page(s)': 'P304',
        'issue': 'P433',
        'author name string': 'P2093',
        'PubMed ID': 'P698',
        'DOI': 'P356',
        'series ordinal': 'P1545',
    }

    # can be expanded to books and such...
    instance_of_items = {'journalArticle': 'Q13442814'}
    descriptions = {'journalArticle': 'scientific journal article'}

    def __init__(self, pmid):
        self.pmid = str(pmid)
        self.meta = {
            'title': None,
            'dois': None,
            'volume': None,
            'issue': None,
            'journal_wdid': None,
            'journal_issn': None,
            'type': None,
            'date': None,
            'pages': None,
            'authors': None,
        }

        self.reference = None
        self.statements = None
        self.wdid = None
        self.citoid = None

    def _valid_pmid(self):
        url = 'https://citoid.wikimedia.org/api?search=https://www.ncbi.nlm.nih.gov/pubmed/{}&format=zotero'.format(
            self.pmid)
        self.citoid = requests.get(url).json()[0]
        return True if self.citoid['itemType'] in self.instance_of_items else False

    def get_metadata(self):

        if not self._valid_pmid():
            raise ValueError("error getting metadata: {}".format(self.pmid))

        # todo: original language of work not in citoid response
        self.meta['title'] = self.citoid['title'][:249]
        self.meta['type'] = self.citoid['itemType']  # instance of. values = {'journalArticle', .. }
        self.meta['journal_issn'] = self.citoid['ISSN']  # published in
        self.meta['date'] = du.parse(self.citoid['date']).strftime('+%Y-%m-%dT%H:%M:%SZ')
        self.meta['authors'] = [x['firstName'] + " " + x['lastName'] for x in self.citoid['creators'] if
                                x['creatorType'] == 'author']

        # optional
        self.meta['volume'] = self.citoid.get('volume', None)
        self.meta['pages'] = self.citoid.get('pages', None)
        self.meta['issue'] = self.citoid.get('issue', None)
        self.meta['dois'] = self.citoid.get('DOI', None)

        # also validate response here
        if self.meta['dois']:
            assert isinstance(self.meta['dois'], str)
        self.meta['journal_wdid'] = prop2qid("P236", self.meta['journal_issn'])
        if not self.meta['journal_wdid']:
            raise ValueError("journal not found: {}".format(self.meta['journal_issn']))

    def make_reference(self):
        stated_in = wdi_core.WDItemID(value='Q180686', prop_nr='P248', is_reference=True)
        pmid = wdi_core.WDString(value=self.pmid, prop_nr='P698', is_reference=True)
        retrieved = wdi_core.WDTime(strftime("+%Y-%m-%dT00:00:00Z", gmtime()), prop_nr='P813', is_reference=True)
        self.reference = [stated_in, pmid, retrieved]

    def make_author_statements(self, ordinals=None):
        """
        ordinals is a list of ordinals to NOT include
        use this to update an existing item that may have authors, so you don't want to
        add the author string for them
        :param ordinals: if None, include all authors
        :return:
        """
        s = []
        for n, author in enumerate(self.meta['authors'], 1):
            if ordinals is not None and n in ordinals:
                continue
            series_ordinal = wdi_core.WDString(str(n), self.PROPS['series ordinal'], is_qualifier=True)
            s.append(wdi_core.WDString(author, self.PROPS['author name string'],
                                       references=[self.reference], qualifiers=[series_ordinal]))
        return s

    def make_statements(self, ordinals=None):
        """

        :param ordinals: passed to self.make_author_statements
        :return:
        """
        self.make_reference()
        s = []

        ### Required statements
        s.append(wdi_core.WDExternalID(self.pmid, self.PROPS['PubMed ID'], references=[self.reference]))
        s.append(wdi_core.WDItemID(self.instance_of_items[self.meta['type']], self.PROPS['instance of'],
                                   references=[self.reference]))
        s.append(wdi_core.WDMonolingualText(self.meta['title'], self.PROPS['title'], references=[self.reference]))
        s.append(wdi_core.WDTime(self.meta['date'], self.PROPS['publication date'], references=[self.reference]))
        s.append(wdi_core.WDItemID(self.meta['journal_wdid'], self.PROPS['published in'], references=[self.reference]))
        s.extend(self.make_author_statements(ordinals=ordinals))

        ### Optional statements
        if self.meta['volume']:
            s.append(wdi_core.WDString(self.meta['volume'], self.PROPS['volume'], references=[self.reference]))
        if self.meta['pages']:
            s.append(wdi_core.WDString(self.meta['pages'], self.PROPS['page(s)'], references=[self.reference]))
        if self.meta['issue']:
            s.append(wdi_core.WDString(self.meta['issue'], self.PROPS['issue'], references=[self.reference]))
        if self.meta['dois']:
            s.append(wdi_core.WDExternalID(self.meta['dois'].upper(), self.PROPS['DOI'], references=[self.reference]))

        self.statements = s

    def update(self, wdid, login):
        # we need the json of the item to grab any existing authors
        url = "https://www.wikidata.org/w/api.php?action=wbgetentities&ids={}&format=json".format(wdid)
        dc = requests.get(url).json()
        claims = dc['entities'][wdid]['claims']
        if 'P50' not in claims:
            ordinals = None
        else:
            ordinals = []
            if not isinstance(claims['P50'], list):
                claims['P50'] = [claims['P50']]
            for claim in claims['P50']:
                assert len(claim['qualifiers']['P1545']) == 1
                ordinals.append(int(claim['qualifiers']['P1545'][0]['datavalue']['value']))

        # validate pmid is on item
        if 'P698' not in claims:
            raise ValueError("unknown pubmed id")
        if len(claims['P698']) != 1:
            raise ValueError("more than one pmid")
        item_pmid = claims['P698'][0]['mainsnak']['datavalue']['value']
        if item_pmid != self.pmid:
            raise ValueError("pmids don't match")

        try:
            self.get_metadata()
        except Exception as e:
            print(e)
            return None
        self.make_statements(ordinals)

        item = wdi_core.WDItemEngine(wd_item_id=wdid, data=self.statements, domain="scientific_article")
        item.set_label(self.meta['title'])
        item.set_description(description=self.descriptions[self.meta['type']], lang='en')
        write_success = try_write(item, self.pmid, 'P698', login)
        if write_success:
            self._cache[self.pmid] = item.wd_item_id
            return item.wd_item_id
        else:
            return None

    def create(self, login=None):
        # create new item
        if login is None:
            raise ValueError("login required to create item")

        try:
            self.get_metadata()
        except Exception as e:
            print(e)
            return None
        self.make_statements()

        item = wdi_core.WDItemEngine(item_name=self.meta['title'], data=self.statements, domain="scientific_article")
        item.set_label(self.meta['title'])
        item.set_description(description=self.descriptions[self.meta['type']], lang='en')
        write_success = try_write(item, self.pmid, 'P698', login)
        if write_success:
            self._cache[self.pmid] = item.wd_item_id
            return item.wd_item_id
        else:
            return None

    def get_or_create(self, login=None):
        """
        Returns wdid if item exists or is created
        returns None if the pmid is not found
        returns None and logs the exception if there is a write failure

        :param login: required to create an item
        :return:
        """
        # check if in local cache
        if self.pmid in self._cache:
            return self._cache[self.pmid]
        # check if exists in wikidata
        wdid = prop2qid('P698', self.pmid)
        if wdid:
            self._cache[self.pmid] = wdid
            return wdid

        # item doesn't exist.
        return self.create(login=login)


def try_write(wd_item, record_id, record_prop, login, edit_summary='', write=True):
    """
    Write a PBB_core item. Log if item was created, updated, or skipped.
    Catch and log all errors.

    :param wd_item: A wikidata item that will be written
    :type wd_item: PBB_Core.WDItemEngine
    :param record_id: An external identifier, to be used for logging
    :type record_id: str
    :param record_prop: Property of the external identifier
    :type record_prop: str
    :param login: PBB_core login instance
    :type login: PBB_login.WDLogin
    :param edit_summary: passed directly to wd_item.write
    :type edit_summary: str
    :param write: If `False`, do not actually perform write. Action will be logged as if write had occured
    :type write: bool
    :return: True if write did not throw an exception, returns the exception otherwise
    """
    if wd_item.require_write:
        if wd_item.create_new_item:
            msg = "CREATE"
        else:
            msg = "UPDATE"
    else:
        msg = "SKIP"

    try:
        if write:
            wd_item.write(login=login, edit_summary=edit_summary)
        wdi_core.WDItemEngine.log("INFO", format_msg(record_id, record_prop, wd_item.wd_item_id, msg))
    except WDApiError as e:
        print(e)
        wdi_core.WDItemEngine.log("ERROR",
                                  format_msg(record_id, record_prop, wd_item.wd_item_id, json.dumps(e.wd_error_msg),
                                             type(e)))
        return e
    except Exception as e:
        print(e)
        wdi_core.WDItemEngine.log("ERROR", format_msg(record_id, record_prop, wd_item.wd_item_id, str(e), type(e)))
        return e

    return True


def format_msg(external_id, external_id_prop, wdid, msg, msg_type=None, delimiter=";"):
    """
    Format message for logging
    :return: str
    """
    fmt = ('{}' + delimiter) * 4 + '{}'  # '{};{};{};{};{}'
    s = fmt.format(external_id, external_id_prop, wdid, msg, msg_type)
    return s


def prop2qid(prop, value):
    """
    Lookup a wikidata item ID from a property and string value. For example, get the item QID for the
     item with the entrez gene id (P351): "899959"
    >>> prop2qid('P351','899959')
    :param prop: property
    :type prop: str
    :param value: value of property
    :type value: str
    :return: wdid as string or None
    """
    arguments = '?item wdt:{} "{}"'.format(prop, value)
    query = 'SELECT * WHERE {{{}}}'.format(arguments)
    results = WDItemEngine.execute_sparql_query(query)
    result = results['results']['bindings']
    if len(result) == 0:
        # not found
        return None
    elif len(result) > 1:
        raise ValueError("More than one wikidata ID found for {} {}: {}".format(prop, value, result))
    else:
        return result[0]['item']['value'].split("/")[-1]


def id_mapper(prop, filters=None, raise_on_duplicate=False, return_as_set=False):
    """
    Get all wikidata ID <-> prop <-> value mappings
    Example: id_mapper("P352") -> { 'A0KH68': 'Q23429083',
                                     'Q5ZWJ4': 'Q22334494',
                                     'Q53WF2': 'Q21766762', .... }
    Optional filters can filter query results.
    Example (get all uniprot to wdid, where taxon is human): id_mapper("P352",(("P703", "Q15978631"),))
    :param prop: wikidata property
    :type prop: str
    :param filters: list of tuples, where the first item is a property, second is a value
    :param raise_on_duplicate: If an ID is found on more than one wikidata item, what action to take?
        This is equivalent to the Distinct values constraint. e.g.: http://tinyurl.com/ztpncyb
        Note that a wikidata item can have more than one ID. This is not checked for
        True: raise ValueError
        False: only one of the values is kept if there are duplicates
    :type raise_on_duplicate: bool
    :param return_as_set: If True, all values in the returned dict will be a set of strings
    :type return_as_set: bool

    If `raise_on_duplicate` is False and `return_as_set` is True, the following can be returned:
    { 'A0KH68': {'Q23429083'}, 'B023F44': {'Q237623', 'Q839742'} }

    :return: dict

    """
    query = "SELECT * WHERE {"
    query += "?item wdt:{} ?id .\n".format(prop)
    if filters:
        for f in filters:
            query += "?item wdt:{} wd:{} .\n".format(f[0], f[1])
    query = query + "}"
    results = WDItemEngine.execute_sparql_query(query)['results']['bindings']
    if not results:
        return None

    ids = [x['id']['value'] for x in results]
    if raise_on_duplicate and len(ids) != len(set(ids)):
        dupe_ids = [x for x, count in Counter(ids).items() if count > 1]
        raise ValueError("duplicate ids: {}".format(
            [(x['id']['value'], x['item']['value']) for x in results if x['id']['value'] in dupe_ids]))

    if return_as_set:
        d = defaultdict(set)
        for x in results:
            d[x['id']['value']].add(x['item']['value'].split('/')[-1])
        return dict(d)
    else:
        return {x['id']['value']: x['item']['value'].split('/')[-1] for x in results}



import os
from setuptools import setup, find_packages
from subprocess import check_output
from subprocess import CalledProcessError

setup_path = os.path.dirname(__file__)

print("setup_path: {}".format(setup_path))


def read(fname):
    return open(os.path.join(setup_path, fname)).read()

MAJOR_VERSION = 0
MINOR_VERSION = 0
REPO_URL = 'https://github.com/sebotic/WikidataIntegrator'

version_file = os.path.join(setup_path, 'VERSION')

try:
    MICRO_VERSION = int(check_output("git rev-list --count master", shell=True).decode('utf-8').strip('\n'))

    f = open(version_file, 'w')
    f.write("{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION))
    f.close()
except:
    if os.path.isfile(version_file):
        f = open(version_file, 'rt')
        version_string = f.readline()
        MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION = version_string.split('.')
    else:
        MICRO_VERSION = 1

# Calculate commit hash
try:
    commit_hash = check_output("git rev-parse HEAD", shell=True).decode('utf-8').strip('\n')
except CalledProcessError:
    commit_hash = ''

f = open('./.git-commit-hash', 'w')
f.write("{}.git\n{}".format(REPO_URL, commit_hash))
f.close()

setup(
    name='wikidataintegrator',
    version="{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION),
    author='Sebastian Burgstaller-Muehlbacher, Greg Stupp, Andra Waagmeester',
    author_email='sburgs@scripps.edu',
    description='Python package for reading and writing to/from Wikidata',
    license='AGPLv3',
    keywords='Wikidata biology chemistry medicine',
    url=REPO_URL,
    download_url='https://github.com/sebotic/WikidataIntegrator/tarball/0.0.325',
    # packages=find_packages(),
    packages=['wikidataintegrator', 'wikidataintegrator.backoff'],
    include_package_data=True,
    # long_description=read('README.md'),
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2.7",
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Intended Audience :: Science/Research",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    install_requires=[
        'requests', 'python-dateutil', 'simplejson'
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Sebastian Burgstaller and Andra Waagmeester'
__licence__ = 'AGPLv3'

domain_incompatibilities = {
    'drugs': [

    ],
    'genes': [

    ],
    'proteins': [

    ],
    'disease ontology': [

    ],
    'interpro': [

    ]
}

# a collection of values for the property 'subclass of' (P279) which are valid for a certain domain.
valid_instances = {
    'drugs': [
        'Q11173',  # chemical compound (only one single, pure chemical compound)
        'Q79529',  # chemical substance (several defined chemical compounds)
        'Q8054',  # protein
        'Q12140',  # pharmaceutical drug
        'Q422248'  # monoclonal antibodies
    ],
    'genes': [
        'Q7187'
    ],
    'proteins': [
        'Q8054'
    ],
    'disease ontology': [

    ],
    'obo': [

    ],
    'interpro': [
        'Q423026',  # active site
        'Q616005',  # binding site
        'Q898273',  # domain
        'Q417841',  # family
        'Q898362',  # ptm
    ],
    'chromosome': [
        'Q186380',
        'Q37748',  # chromosome
        'Q667636'
    ],
    'scientific_article': [

    ]
}

wd_properties = {
    'P715': {
        'datatype': 'string',
        'name': 'Drugbank ID',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P279': {
        'datatype': 'item',
        'name': 'subclass of',
        'domain': ['generic'],
        'core_id': 'False'
    },
    'P31': {
        'datatype': 'item',
        'name': 'instance of',
        'domain': ['generic', 'genomes', 'drugs'],
        'core_id': 'False'
    },
    'P636': {
        'datatype': 'item',
        'name': 'route of administration',
        'domain': ['drugs'],
        'core_id': 'False'
    },
    'P267': {
        'datatype': 'string',
        'name': 'ATC code',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P231': {
        'datatype': 'string',
        'name': 'CAS registry number',
        'domain': ['drugs'],
        'core_id': 'False'
    },
    'P486': {
        'datatype': 'string',
        'name': 'MeSH ID',
        'domain': ['drugs', 'diseases'],
        'core_id': 'True'
    },
    'P672': {
        'datatype': 'string',
        'name': 'MeSH Code',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P662': {
        'datatype': 'string',
        'name': 'PubChem ID (CID)',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P661': {
        'datatype': 'string',
        'name': 'ChemSpider ID',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P652': {
        'datatype': 'string',
        'name': 'UNII',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P665': {
        'datatype': 'string',
        'name': 'KEGG ID',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P683': {
        'datatype': 'string',
        'name': 'ChEBI ID',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P274': {
        'datatype': 'string',
        'name': 'chemical formula',
        'domain': ['drugs'],
        'core_id': 'False'
    },
    'P592': {
        'datatype': 'string',
        'name': 'ChEMBL ID',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P233': {
        'datatype': 'string',
        'name': 'SMILES',
        'domain': ['drugs'],
        'core_id': 'False'
    },
    'P2017': {
        'datatype': 'string',
        'name': 'Isomeric SMILES',
        'domain': ['drugs'],
        'core_id': 'False'
    },
    'P234': {
        'datatype': 'string',
        'name': 'InChI',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P235': {
        'datatype': 'string',
        'name': 'InChIKey',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P1805': {
        'datatype': 'string',
        'name': 'Word Health Organisation International Nonproprietary Name',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P657': {
        'datatype': 'string',
        'name': 'RTECS Number',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P595': {
        'datatype': 'string',
        'name': 'IUPHAR ID',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P2115': {
        'datatype': 'string',
        'name': 'NDF-RT NUI',
        'domain': ['drugs'],
        'core_id': 'True'
    },
    'P699': {
        'datatype': 'string',
        'name': 'Disease Ontology ID',
        'domain': ['diseases'],
        'core_id': 'True'
    },
    'P1550': {
        'datatype': 'string',
        'name': 'Orphanet ID',
        'domain': ['diseases'],
        'core_id': 'True'
    },
    'P494': {
        'datatype': 'string',
        'name': 'ICD-10',
        'domain': ['diseases'],
        'core_id': 'False'
    },
    'P493': {
        'datatype': 'string',
        'name': 'ICD-9',
        'domain': ['diseases'],
        'core_id': 'False'
    },
    'P492': {
        'datatype': 'string',
        'name': 'OMIM',
        'domain': ['diseases'],
        'core_id': 'True'
    },
    'P1395': {
        'datatype': 'string',
        'name': 'National Cancer Thesaurus ID',
        'domain': ['diseases'],
        'core_id': 'True'
    },
    'P1748': {
        'datatype': 'string',
        'name': 'NCI Thesaurus ID',
        'domain': ['diseases'],
        'core_id': 'True'
    },
    'P351': {
        'datatype': 'string',
        'name': 'Entrez Gene ID',
        'domain': ['genes'],
        'core_id': 'True'
    },
    'P2393': {
        'datatype': 'string',
        'name': 'NCBI Locus tag',
        'domain': ['genes'],
        'core_id': 'True'
    },
    'P3406': {
        'datatype': 'string',
        'name': 'Saccharomyces Genome Database ID',
        'domain': ['genes'],
        'core_id': 'False'
    },
    'P703': {
        'datatype': 'item',
        'name': 'found in taxon',
        'domain': ['genes', 'chromosome'],
        'core_id': 'False'
    },
    'P594': {
        'datatype': 'string',
        'name': 'Ensembl Gene ID',
        'domain': ['genes'],
        'core_id': 'False'
    },
    'P704': {
        'datatype': 'string',
        'name': 'Ensembl Transcript ID',
        'domain': ['genes'],
        'core_id': 'False'
    },
    'P353': {
        'datatype': 'string',
        'name': 'Human Gene symbol',
        'domain': ['genes'],
        'core_id': 'True'
    },
    'P354': {
        'datatype': 'string',
        'name': 'HGNC symbol',
        'domain': ['genes'],
        'core_id': 'True'
    },
    'P593': {
        'datatype': 'string',
        'name': 'homologene id',
        'domain': ['genes'],
        'core_id': 'False'
    },
    'P639': {
        'datatype': 'string',
        'name': 'RefSeq RNA ID',
        'domain': ['genes'],
        'core_id': 'True'
    },
    'P1057': {
        'datatype': 'item',
        'name': 'chromosome',
        'domain': ['genes'],
        'core_id': 'False'
    },
    'P684': {
        'datatype': 'item',
        'name': 'ortholog',
        'domain': ['genes'],
        'core_id': 'False'
    },
    'P638': {
        'datatype': 'string',
        'name': 'PDB ID',
        'domain': ['proteins'],
        'core_id': 'False'
    },
    'P637': {
        'datatype': 'string',
        'name': 'Refseq Protein ID',
        'domain': ['proteins'],
        'core_id': 'False'
    },
    'P352': {
        'datatype': 'string',
        'name': 'Uniprot ID',
        'domain': ['proteins'],
        'core_id': 'True'
    },
    'P591': {
        'datatype': 'string',
        'name': 'EC Number',
        'domain': ['proteins'],
        'core_id': 'False'
    },
    'P705': {
        'datatype': 'string',
        'name': 'Ensembl Protein ID',
        'domain': ['proteins'],
        'core_id': 'False'
    },
    'P702': {
        'datatype': 'item',
        'name': 'Encoded By',
        'domain': ['proteins'],
        'core_id': 'False'
    },
    'P18': {
        'datatype': 'url',
        'name': 'Protein Structure Image',
        'domain': ['proteins'],
        'core_id': 'True'
    },
    'P671': {
        'datatype': 'url',
        'name': 'MGI',
        'domain': ['genes'],
        'core_id': 'True'
    },
    'P644': {
        'datatype': 'string',
        'name': 'Genomic start position',
        'domain': ['genes'],
        'core_id': 'False'
    },
    'P645': {
        'datatype': 'string',
        'name': 'Genomic end position',
        'domain': ['genes'],
        'core_id': 'False'
    },
    'P688': {
        'datatype': 'item',
        'name': 'encodes',
        'domain': ['genes', 'proteins'],
        'core_id': 'False'
    },
    'P225': {
        'datatype': 'string',
        'name': 'taxon name',
        'domain': ['genomes'],
        'core_id': 'False'
    },
    'P685': {
        'datatype': 'string',
        'name': 'NCBI Taxonomy ID',
        'domain': ['genomes'],
        'core_id': 'True'
    },
    'P171': {
        'datatype': 'item',
        'name': 'parent taxon',
        'domain': ['genomes'],
        'core_id': 'False'
    },
    'P1065': {
        'datatype': 'url',
        'name': 'Archive url',
        'domain': ['genomes'],
        'core_id': 'True'
    },
    'P856': {
        'datatype': 'url',
        'name': 'offical website',
        'domain': ['genomes'],
        'core_id': 'False'
    },
    'P680': {
        'datatype': 'item',
        'name': 'molecular function',
        'domain': ['proteins'],
        'core_id': 'False'
    },
    'P681': {
        'datatype': 'item',
        'name': 'cell component',
        'domain': ['proteins'],
        'core_id': 'False'
    },
    'P682': {
        'datatype': 'item',
        'name': 'biological process',
        'domain': ['proteins'],
        'core_id': 'False'
    },
    'P1554': {
        'datatype': 'item',
        'name': 'uberon id',
        'domain': ['anatomical_structure', 'obo'],
        'core_id': 'False'
    },
    'P1709': {
        'datatype': 'item',
        'name': 'equivalent class',
        'domain': ['genes', 'proteins', 'diseases', 'drugs', 'anatomical_structure'],
        'core_id': 'False'
    },
    'P686': {
        'datatype': 'item',
        'name': 'Gene Ontology ID',
        'domain': ['obo'],
        'core_id': 'True'
    },
    'P2926': {
        'datatype': 'string',
        'name': 'InterPro ID',
        'domain': ['interpro'],
        'core_id': 'True'
    },
    'P2870': {
        'datatype': 'string',
        'name': 'miRBase pre-miRNA ID',
        'domain': ['microRNAs'],
        'core_id': 'True'
    },
    'P2871': {
        'datatype': 'string',
        'name': 'miRBase mature miRNA ID',
        'domain': ['microRNAs'],
        'core_id': 'True'
    },
    'P2646': {
        'datatype': 'string',
        'name': 'mirTarBase ID',
        'domain': ['microRNAs'],
        'core_id': 'True'
    },
    'P2249': {
        'datatype': 'string',
        'name': 'Refseq Genome ID',
        'domain': ['chromosome'],
        'core_id': 'True'
    },
    'P698': {
        'datatype': 'string',
        'name': 'PubMed ID',
        'domain': ['scientific_article'],
        'core_id': 'True'
    }
}
